import torch
import torch.nn as nn

from .conditioners import Conditioner, StrNNConditioner, GNFConditioner, \
    MADEConditioner
from .normalizers import Normalizer, AffineNormalizer, MonotonicNormalizer

from ..normalizing_flow import NormalizingFlow, NormalizingFlowFactory
from ..config_constants import *

from ...models import TTuple


class AutoregressiveFlowStep(NormalizingFlow):
    """A single normalizing flow step which can be chained together."""

    def __init__(self, conditioner: Conditioner, normalizer: Normalizer):
        """Initialize an autoregressive flow step.

        Args:
            conditioner: Flow conditioner.
            normalizer: Flow normalizer.
        """
        super().__init__()
        self.conditioner = conditioner
        self.normalizer = normalizer

    def forward(self, x: torch.Tensor) -> TTuple:
        """Compute forward pass of flow step.

        Args:
            x: Input data.
        Returns:
            Transformed data, and jacobian determinant.
        """
        params = self.conditioner(x)
        z, jac = self.normalizer(x, params)
        return z, torch.log(jac).sum(1)

    def invert(self, z: torch.Tensor) -> torch.Tensor:
        """Perform inverse transformation from latent to observed space.

        Args:
            z: Input data from latent space.
        Returns:
            Transformed data.
        """
        x = torch.zeros_like(z)

        for i in range(self.conditioner.input_dim + 1):
            params = self.conditioner(x)
            x_prev = x
            x = self.normalizer.invert(z, params)
            if torch.norm(x - x_prev) == 0.:
                break
        return x


class AutoregressiveFlow(NormalizingFlow):
    """Implements a chain of autoregressive flow steps."""

    def __init__(self, steps: list[AutoregressiveFlowStep], permute: bool):
        """Initialize AutoregressiveFlow.

        Args:
            steps: List of AutoregressiveFlowSteps to be composed.
            permute: Flips the variable order between flow steps.
        """
        super().__init__()
        self.permute = permute
        self.steps = nn.ModuleList(steps)

    def forward(self, x: torch.Tensor) -> TTuple:
        """Compute forward pass through chained steps.

        Args:
            x: Input data.

        Returns:
            Transformed data and jacobian determinant.
        """
        jac_tot = torch.zeros(x.shape[0]).to(x)

        inv_idx = torch.arange(x.shape[1] - 1, -1, -1).long()
        for step in self.steps:
            z, jac = step(x)
            jac_tot += jac

            if self.permute:
                x = z[:, inv_idx]
            else:
                x = z

        return z, jac_tot

    def invert(self, z: torch.Tensor) -> torch.Tensor:
        """Inverted transform of composed flow steps.

        Args:
            z: Input data from latent space.
        Returns:
            Transformed data.
        """
        for step in range(1, len(self.steps) + 1):
            x = self.steps[-step].invert(z)

            if self.permute:
                inv_idx = torch.arange(z.shape[1] - 1, -1, -1).long()
                z = x[:, inv_idx]
            else:
                z = x

        return x

    def invert_gnf(self, z: torch.Tensor) -> torch.Tensor:
        """Perform inversion according to the GNF offical repo.

        Importantly, variables are NOT re-permuted. This leads to
        poor sample quality. The order of flow steps is also
        incorrect due to an indexing error.

        Args:
            z: Input data from latent space.
        Returns:
            Transformed data.
        """
        for step in range(len(self.steps)):
            z = self.steps[-step].invert(z)

        return z


class AutoregressiveFlowFactory(NormalizingFlowFactory):
    """Constructs AutoregressiveFlow."""

    def __init__(self, config: dict):
        """Initialize AutoregressiveFlowFactory.

        Args:
            config: Dictionary of model initialization attributes.
        """
        super().__init__(config)
        self.parse_config(config)

    def parse_config(self, config: dict):
        """Parse config and stores relevant attributes."""
        self.input_dim = config[INPUT_DIM]
        self.adj = config[ADJ]
        self.flow_steps = config[FLOW_STEPS]
        self.flow_permute = config[FLOW_PERMUTE]

        self.cond_type = config[COND_TYPE]
        self.norm_type = config[NORM_TYPE]

        self.cond_hid = config[COND_HID]
        self.cond_act = config[COND_ACT]

        self.cond_out = config[N_PARAM_PER_VAR]

        if self.cond_type == "strnn":
            self.opt_type = config[OPT_TYPE]
            self.opt_args = config[OPT_ARGS]
        elif self.cond_type == "gnf":
            self.gnf_hot = config[GNF_HOT]
        elif self.cond_type == "made":
            pass
        else:
            raise ValueError("Unknown conditioner type.")

        if self.norm_type == "affine":
            assert self.cond_out == 2
        elif self.norm_type == "umnn":
            self.umnn_int_hid = config[UMNN_INT_HID]
            self.umnn_step = config[UMNN_INT_STEP]
            self.umnn_solver = config[UMNN_INT_SOLVER]
        else:
            raise ValueError("Unknown normalizer type.")

    def _build_flow_step(self) -> AutoregressiveFlowStep:
        """Construct a single autoregressive flow step according to config.

        Returns:
            Initialized AutoregressiveFlowStep.
        """
        conditioner: Conditioner
        normalizer: Normalizer

        if self.cond_type == "strnn":
            conditioner = StrNNConditioner(
                self.input_dim,
                self.cond_hid,
                self.cond_out,
                self.cond_act,
                self.adj,
                self.opt_type,
                self.opt_args
            )
        elif self.cond_type == "made":
            conditioner = MADEConditioner(
                self.input_dim,
                self.cond_hid,
                self.cond_out,
                self.cond_act
            )
        elif self.cond_type == "gnf":
            conditioner = GNFConditioner(
                self.input_dim,
                self.cond_hid,
                self.cond_out,
                self.cond_act,
                self.gnf_hot,
                self.adj
            )
        else:
            raise ValueError("Unknown conditioner type.")

        if self.norm_type == "affine":
            normalizer = AffineNormalizer()
        elif self.norm_type == "umnn":
            normalizer = MonotonicNormalizer(
                self.umnn_int_hid,
                self.cond_out,
                self.umnn_step,
                self.umnn_solver
            )
        else:
            raise ValueError("Unknown normalizer type.")

        return AutoregressiveFlowStep(conditioner, normalizer)

    def _build_flow(self) -> AutoregressiveFlow:
        """Build AutoregressiveFlow out of composed flow steps.

        Returns:
            Composition of autoregressive flow steps.
        """
        flow_steps = [self._build_flow_step() for _ in range(self.flow_steps)]

        flow = AutoregressiveFlow(flow_steps, self.flow_permute)
        return flow



from abc import ABC, abstractmethod

import numpy as np


def check_masks(masks: list[np.ndarray], adjacency: np.ndarray) -> bool:
    """Check whether weight masks respect prescribed adjacency matrix.

    Given a set of masks, [M1, M2, ..., Mk], check if the matrix product
    (M1 * M2 * ... * Mk).T = Mk.T * ... * M2.T * M1.T
    respects the provided adjacency structure.

    Args:
        masks: List of masks [M1, M2, ..., Mk]
        adjacency: Adjacency matrix of shape (n_output x n_input)

    Returns:
        True or False depending on validity of weight masks.
    """
    mask_prod = masks[0]
    for i in range(1, len(masks)):
        mask_prod = mask_prod @ masks[i]
    mask_prod = mask_prod.T

    constraint = (mask_prod > 0.0001) * 1. - adjacency

    return not np.any(constraint != 0.)


class AdjacencyFactorizer(ABC):
    """Interface for adjacency factorization algorithms."""

    @abstractmethod
    def __init__(self, adjacency: np.ndarray, opt_args: dict | None = None):
        """Abstract initialization template."""
        pass

    @abstractmethod
    def factorize(
        self,
        hidden_sizes: tuple[int, ...]
    ) -> list[np.ndarray]:
        """Abstract factorization method."""
        pass



from abc import ABCMeta, abstractmethod

import numpy as np

import torch
import torch.nn as nn

from strnn.models.strNN import StrNN
from ..model_utils import NONLINEARITIES

from ...models import Array_like


class ODENet(nn.Module, metaclass=ABCMeta):
    """Interface for networks used to model ODE dynamics."""

    @abstractmethod
    def forward(self, t: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """Abstract method for ODENet forward pass."""
        pass


class WeilbachSparseLinear(nn.Module):
    """Concatenated squash linear layer.

    See: http://proceedings.mlr.press/v108/weilbach20a/weilbach20a.pdf.
    Implementation adapted from https://github.com/plai-group/daphne.
    """

    _weight_mask: torch.Tensor

    def __init__(self, dim_in: int, dim_out: int, adj_mat: Array_like):
        """Initialize a sparse concat squash layer as in Weilbach et al. 2020.

        Args:
            dim_in: Input dimension.
            dim_out: Output dimension.
            adj_mat: 2D binary adjacency matrix.
        """
        super().__init__()

        self.dim_in = dim_in
        self.dim_out = dim_out
        self.adj_mat = torch.Tensor(adj_mat)

        self.register_buffer("_weight_mask", torch.zeros([dim_out, dim_in]))
        self._weight_mask[:adj_mat.shape[0], :adj_mat.shape[1]] = self.adj_mat

        lin = nn.Linear(dim_in, dim_out)
        self._weights = lin.weight
        self._bias = lin.bias

        self._hyper_bias = nn.Linear(1, dim_out, bias=False)
        self._hyper_gate = nn.Linear(1, dim_out)

    def forward(self, t: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """Compute forward pass.

        Args:
            t: Time of integration.
            x: Data input.

        Returns:
            Predicted dynamics.
        """
        w = torch.mul(self._weight_mask, self._weights)
        res = torch.addmm(self._bias, x, w.transpose(0, 1))

        hyper_bias = self._hyper_bias(t.view(1, 1))
        return res * torch.sigmoid(self._hyper_gate(t.view(1, 1))) + hyper_bias


class WeilbachSparseODENet(ODENet):
    """Implements the sparse ODENet described below.

    http://proceedings.mlr.press/v108/weilbach20a/weilbach20a.pdf

    Code adapted from: https://github.com/plai-group/daphne.

    As linear layers enforce independencies by directly multiplying the
    adjacency matrix and offers no factorization, multiple consecutive hidden
    layers are not possible, nor are hidden layers that are not the size of
    the adjacency matrix.
    """

    def __init__(
        self,
        input_dim: int,
        num_layer: int,
        act_type: str,
        adj_mat: Array_like
    ):
        """Initialize sparse ODENet from Weilbach et al. 2020.

        Args:
            input_dim: Input dimension.
            num_layer: Number of stacked WeilbachSparseLinear layers.
            act_type: Type of activation function used between layers.
            adj_mat: 2D Binary adjacency matrix.
        """
        super().__init__()

        adj_mat = torch.Tensor(adj_mat)
        self.adj_mat = adj_mat

        layers = [WeilbachSparseLinear(input_dim + 1, input_dim, adj_mat)]

        for _ in range(num_layer):
            layers.append(WeilbachSparseLinear(input_dim, input_dim, adj_mat))

        self.layers = nn.ModuleList(layers)

        activation_fn = NONLINEARITIES[act_type]
        activation_fns = [activation_fn for _ in range(num_layer)]
        self.activation_fns = nn.ModuleList(activation_fns)

    def forward(self, t: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """Compute forward pass.

        Args:
            t: Time of integration.
            x: Data input.

        Returns:
            Predicted dynamics.
        """
        batch_dim = x.shape[0]
        dx = torch.cat([x, t * torch.ones([batch_dim, 1]).to(x)], dim=1)

        for i, layer in enumerate(self.layers):
            # if not last layer, use nonlinearity
            if i < len(self.layers) - 1:
                acti = layer(t, dx)
                if i == 0:
                    dx = self.activation_fns[i](acti)
                else:
                    dx = self.activation_fns[i](acti) + dx
            else:
                dx = layer(t, dx)
        return dx


class IgnoreLinear(nn.Module):
    """Fully connected layer for use in ODE Net.

    Code copied from FFJORD: https://github.com/rtqichen/ffjord.
    """

    def __init__(self, dim_in: int, dim_out: int):
        """Initialize IgnoreLinear layer.

        Args:
            dim_in: Input dimension.
            dim_out: Output dimension.
        """
        super().__init__()
        self._layer = nn.Linear(dim_in, dim_out)

    def forward(self, t: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """Compute forward pass using dummy layer.

        Args:
            t: Time of integration, unused.
            x: Data values.

        Returns:
            Predicted dynamics.
        """
        return self._layer(x)


class FCODEnet(ODENet):
    """Fully connected ODENet.

    Code modified from FFJORD: https://github.com/rtqichen/ffjord.

    Although better FFJORD ODENets exists, this fully connection layer is the
    best negative control for against the StrODENet and WeilbachSparseODENet.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dims: tuple[int],
        act_type: str
    ):
        """Initialize a fully connected ODENet.

        Args:
            input_shape: Input dimension.
            hidden_dims: List containing widths of hidden dimensions.
            act_type: Activation function between layers.

        Returns:
            Fully connected NN for use in Neural ODE.
        """
        super().__init__()

        layers = []
        activation_fns = []
        hidden_shape = input_dim

        for dim_out in list(hidden_dims) + [input_dim]:
            layers.append(IgnoreLinear(hidden_shape, dim_out))
            activation_fns.append(NONLINEARITIES[act_type])

            hidden_shape = dim_out

        self.layers = nn.ModuleList(layers)
        self.activation_fns = nn.ModuleList(activation_fns[:-1])

    def forward(self, t: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Compute forward pass.

        Args:
            t: Time of integration.
            x: Data input.

        Returns:
            Predicted dynamics.
        """
        dx = y
        for i, layer in enumerate(self.layers):
            dx = layer(t, dx)
            # if not last layer, use nonlinearity
            if i < len(self.layers) - 1:
                dx = self.activation_fns[i](dx)
        return dx


class StrODENet(ODENet):
    """Wraps StrNN model for use as an ODE dynamic function."""

    def __init__(
        self,
        input_dim: int,
        hidden_dim: tuple[int],
        activation: str,
        opt_type: str,
        opt_args: dict,
        adjacency: np.ndarray
    ):
        """Initialize StrODENet.

        Args:
            input_dim: Input dimension.
            hidden_dim: List of StrNN hidden widths per layer.
            activation: Activation function between layers.
            opt_type: Factorizer used by StrNN.
            opt_args: Arguments to pass to factorizer.
            adjacency: Adjacency matrix.
        """
        super().__init__()
        self.net = StrNN(
            input_dim,
            hidden_dim,
            input_dim,
            opt_type,
            opt_args,
            None,
            adjacency,
            activation
        )

    def forward(self, t: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """Compute dynamics using StrNN.

        Currently uses an autonomous function.

        Args:
            t: Time of integration, unused.
            x: Data values.

        Returns:
            Predicted dynamics.
        """
        return self.net(x)



from .greedy import GreedyFactorizer
from .made import MADEFactorizer
from .zuko import ZukoFactorizer
from .factorizer import check_masks

__all__ = ["GreedyFactorizer", "MADEFactorizer", "ZukoFactorizer"]
__all__ += ["check_masks"]



import numpy as np
from sklearn.model_selection import train_test_split

import torch
import torch.nn as nn
import torch.optim as optim

from torch.distributions import Distribution, Laplace, Uniform, Normal
from torch.distributions import TransformedDistribution, SigmoidTransform
from torch.utils.data import DataLoader

from data.causal_sem import CustomSyntheticDatasetDensity

from strnn.models.discrete_flows import AutoregressiveFlow
from strnn.models.discrete_flows import AutoregressiveFlowFactory

from argparse import Namespace


class CausalAutoregressiveFlowWithPrior(nn.Module):
    """Autoregressive flow generated from user specified prior distribution.

    Can be used to define both the Causal StrAF and the baseline CAREFL model,
    which is a Causal StrAF with a fully autoregressive adjacency matrix.

    The CAREFL model developed in Causal Autoregressive Flows,
    by Khemakhem et al. (2021) is available at:
        https://arxiv.org/abs/2011.02268

    The CAREFL and the Causal StrAF can be used to find causal direction
    between pairs of (multivariate) random variables, or to perform
    interventions and answer counterfactual queries.

    Code to perform intervention and counterfactual queries have been adapted
    from CAREFL, but several bugs have been fixed in our implementation. These
    fixes have been noted in the correspond methods.
    """

    def __init__(self, prior: Distribution, flow: AutoregressiveFlow):
        """Initialize an autoregressive flow with a flexible prior.

        Args:
            prior:
                PyTorch distribution describing latent-space prior. Note that
                the prior must match the dimensionality of the latent space,
                as other methods use the prior to generate samples.
            flow:
                AR flow used to perform invertible transformations.
        """
        super().__init__()
        self.prior = prior
        self.flow = flow

    def forward(
        self,
        x: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Transform data-space samples to latent-space and get likelihood.

        Args:
            Data-space samples to normalize of shape (B x D)

        Returns:
            Normalized samples, log prior probability, and Jacobian determinant
        """
        zs, log_det = self.flow.forward(x)
        prior_logprob = self.prior.log_prob(zs).view(x.size(0), -1).sum(1)
        return zs, prior_logprob, log_det

    def backward(self, z: torch.Tensor) -> torch.Tensor:
        """Transform latent-space samples to data-space using flow.

        Args:
            z: Samples in latent-space of shape (B x D)

        Returns:
            Samples in data-space after flow transformation of shape (B x D)
        """
        xs = self.flow.invert(z)
        return xs

    def sample(self, num_samples: int) -> torch.Tensor:
        """Generate data-space samples using inverse flow transformation.

        Args:
            num_samples: Number of samples to generate

        Returns:
            Samples in data space generated from flow distribution
        """
        z = self.prior.sample(torch.Size((num_samples,)))
        xs = self.flow.invert(z)
        return xs

    def log_likelihood(self, x: torch.Tensor) -> np.ndarray:
        """Compute log likelihood of samples in data-space.

        Args:
            x: Data sample of shape (B x D)

        Returns:
            Log likelihood of data samples
        """
        _, prior_logprob, log_det = self.forward(x)
        return (prior_logprob + log_det).detach().cpu().numpy()

    def predict_intervention(
        self,
        x0_val: float,
        n_samples: int = 100,
        iidx: int = 0
    ) -> np.ndarray:
        """Predict the value of x given an intervention do(x_iidx = x0_val).

        As explained in Appendix E.1 of the CAREFL paper, we do not need to
        invert the flow to perform interventions, but doing so provides a
        simpler and more efficient algorithm. Thus, this implementation will
        follow Algorithm 2 of Appendix E.1.

        However, the CAREFL implementation is actually incorrect, and has
        been fixed here. The intervention proceeds as below:

        1) sample noises z from base distribution
        2) pass z through flow to get initial samples for x
        3) invert flow to find corresponding entry for z_iidx at x_iidx=x0_val
        4) pass z again through flow to get samples for x

        Args:
            x0_val: Interventional value
            n_samples: Number of samples used to estimate interventional effect
            iidx: Index of variable to intervene on

        Return:
            Estimate of interventional effect
        """
        # Sample from prior and ensure z_intervention_index = z_int
        z = self.prior.sample(torch.Size((n_samples,)))

        # Pass z through flow to get initial samples for x
        x_init = self.backward(z)

        # Invert flow to infer value of latent corresponding to
        # interventional variable
        x_init[:, iidx] = x0_val
        z_int = self.forward(x_init)[0][:, iidx]
        z[:, iidx] = z_int

        # Propagate the latent sample through flow
        x = self.backward(z)

        # Sanity check: check x_intervention_index == x0_val
        assert (torch.abs(x[:, iidx] - x0_val) < 1e-3).all()

        int_est = x.mean(0).reshape((1, z.shape[1]))
        return int_est.detach().cpu().numpy()

    def predict_counterfactual(
        self,
        x_obs: torch.Tensor | np.ndarray,
        cf_val: float,
        iidx: int = 0
    ) -> np.ndarray:
        """Predict counterfactual of setting dimension iidx of x_obs to cf_val.

        For clarity, we restrict counterfacutal queries to a single example.
        However, this method can easily be extended to accept an array of
        counterfactual values instead.

        Given observation x_obs we estimate the counterfactual of setting
        x_obs[intervention_index] = cf_val

        This proceeds in 3 steps:
         1) abduction - pass-forward through flow to infer latents for x_obs
         2) action - pass-forward again for latent associated with cf_val
         3) prediction - backward pass through the flow

        Args:
            x_obs: Observed data sample to perform counterfactual query.
            cf_val: Counterfactual value.
            iidx: Feature dimension of counterfactual value.

        Returns:
            Counterfactual of observed data sample
        """
        if x_obs.shape[0] != 1:
            raise ValueError("Counterfactual only accepts single example.")

        if isinstance(x_obs, np.ndarray):
            x_obs = torch.from_numpy(x_obs)

        # abduction:
        z_obs = self.forward(x_obs)[0]
        # action (get latent variable value under counterfactual)
        x_cf = torch.clone(x_obs)
        x_cf[0, iidx] = cf_val
        z_cf_val = self.forward(x_cf)[0][0, iidx]
        z_obs[0, iidx] = z_cf_val

        # prediction (pass through the flow):
        x_post_cf = self.backward(z_obs)
        return x_post_cf.detach().cpu().numpy()


class CausalARFlowTrainer:
    """Wrapper around the Causal StrAF / CAREFL to allow easier training."""

    def __init__(self, config: Namespace):
        """Initialize a causal flow trainer.

        Args:
            config: Dictionary of model arguments. See below for parameters

            config.arch_config: Config passed to AutoregressiveFlow init
            config.device: CUDA device used for training
            config.flow.prior_dist: Prior distribution used for likelihood
            config.optim.adam_beta: Adam optimizer beta_1 coefficient
            config.optim.amsgrad: Whether to use amsgrad in Adam optimizer
            config.optim.lr: Optimization learning rate
            config.optim.scheduler: Whether a learning rate scheduler is used
            config.optim.weight_decay: Adam weight decay value
            config.training.batch_size: Batch size used in training
            config.training.early_stop_patience: Epochs before early stopping
            config.training.epoch: Number of training epochs
            config.training.seed: Random seed used for training
            config.training.split: Training split ratio
            config.training.verbose: Whether training logs are printed
        """
        self.config: Namespace = config
        self.arch_config: dict = config.arch_config
        self.prior_dist: Distribution = config.flow.prior_dist

        self.device: torch.device = config.device

        self.adam_beta: float = config.optim.beta1
        self.amsgrad: bool = config.optim.amsgrad
        self.lr: float = config.optim.lr
        self.scheduler: bool = config.optim.scheduler
        self.weight_decay: float = config.optim.weight_decay

        self.batch_size: int = config.training.batch_size
        self.epochs: int = config.training.epochs
        self.seed: int = config.training.seed
        self.split_ratio: float = config.training.split
        self.verbose: bool = config.training.verbose
        self.stopping_patience: int = config.training.early_stop_patience

        self.dim: int | None = None
        self.flow: CausalAutoregressiveFlowWithPrior | None = None

    def fit_to_sem(self, data: np.ndarray) -> tuple[list[float], list[float]]:
        """Fits SEM assuming data columns follow the causal ordering.

        Args:
            data:
                A dataset whose columns are ordered following the causal
                ordering $pi$. The causal ordering can be a result of causal
                discovery algorithm, or expert judgement.
        """
        train_dl, val_dl = self._get_datasets(data)

        torch.manual_seed(self.seed)
        self.flow = self._get_flow_arch()

        train_losses, val_losses = self._train(self.flow, train_dl, val_dl)

        return train_losses, val_losses

    def _get_datasets(self, data: np.ndarray) -> tuple[DataLoader, DataLoader]:
        """Get train/val dataset splits from overall design matrix.

        Args:
            dataset: Observed samples

        Returns:
            Training and validation dataloaders
        """
        assert isinstance(data, np.ndarray)

        self.dim = data.shape[-1]

        train_np, val_np = train_test_split(data, train_size=self.split_ratio)

        train = CustomSyntheticDatasetDensity(train_np.astype(np.float32))
        val = CustomSyntheticDatasetDensity(val_np.astype(np.float32))

        train_dl = DataLoader(train, shuffle=True, batch_size=self.batch_size)
        val_dl = DataLoader(val, shuffle=False, batch_size=self.batch_size)

        return train_dl, val_dl

    def _get_flow_arch(self) -> CausalAutoregressiveFlowWithPrior:
        """Return a normalizing flow according to the config file.

        Returns:
            Initialized Causal Autoregressive Flow (Causal StrAF or CAREFL).
        """
        if self.dim is None:
            raise RuntimeError("self._get_datasets must be called first.")

        prior: Distribution | None = None

        # Setup priors
        if self.prior_dist == 'laplace':
            prior = Laplace(
                torch.zeros(self.dim).to(self.device),
                torch.ones(self.dim).to(self.device)
            )
        elif self.prior_dist == 'normal':
            prior = Normal(
                torch.zeros(self.dim).to(self.device),
                torch.ones(self.dim).to(self.device)
            )
        else:
            prior = TransformedDistribution(
                Uniform(
                    torch.zeros(self.dim).to(self.device),
                    torch.ones(self.dim).to(self.device)
                ),
                SigmoidTransform().inv
            )

        # Build flow model
        model_factory = AutoregressiveFlowFactory(self.arch_config)
        flow = model_factory.build_flow()

        assert isinstance(flow, AutoregressiveFlow)

        return CausalAutoregressiveFlowWithPrior(prior, flow).to(self.device)

    def _get_optimizer(
        self,
        model: CausalAutoregressiveFlowWithPrior,
    ) -> tuple[optim.Optimizer, optim.lr_scheduler.ReduceLROnPlateau | None]:
        """Return an optimizer according to the config file.

        Args:
            model: Model to optimize

        Returns:
            Initialized Adam optimizer, and optionally a LR scheduler
        """
        optimizer = optim.Adam(
            model.parameters(),
            lr=self.lr,
            weight_decay=self.weight_decay,
            betas=(self.adam_beta, 0.999),
            amsgrad=self.amsgrad
        )

        if self.scheduler:
            scheduler = optim.lr_scheduler.ReduceLROnPlateau(
                optimizer,
                factor=0.1,
                patience=5,
                verbose=self.verbose
            )
        else:
            scheduler = None

        return optimizer, scheduler

    def _train(
        self,
        flow: CausalAutoregressiveFlowWithPrior,
        train_dl: DataLoader,
        val_dl: DataLoader
    ) -> tuple[list, list]:
        """Train causal autoregressive flow.

        Implements early stopping if the stopping_patience attribute is not
        zero. This halts training if validation loss does not improve the
        number of epochs specified by stopping_patience. The model is then set
        to the weights which yielded lowest loss.

        Args:
            flow: Causal autoregressive flow to train
            train_dl: Train dataloader
            val_dl: Validation dataloader

        Returns:
            Train and validation losses throughout training.
        """
        optimizer, scheduler = self._get_optimizer(flow)

        flow.train()

        train_losses = []
        val_losses = []

        best_val_loss = None
        best_weights = flow.state_dict()
        patience_counter = 0

        for e in range(self.epochs):
            epoch_t_loss = []
            epoch_v_loss = []

            for train_batch in train_dl:
                optimizer.zero_grad()
                train_batch = train_batch.to(self.device)

                # compute loss
                _, prior_logprob, log_det = flow(train_batch)
                train_loss = -torch.sum(prior_logprob + log_det)
                epoch_t_loss.append(train_loss.item())

                # optimize
                train_loss.backward()
                optimizer.step()

            with torch.no_grad():
                flow.eval()

                for val_batch in val_dl:
                    val_batch = val_batch.to(self.device)

                    # compute loss
                    _, prior_logprob, log_det = flow(val_batch)
                    val_loss = -torch.sum(prior_logprob + log_det)
                    epoch_v_loss.append(val_loss.item())

                flow.train()

            epoch_t_mean = np.mean(epoch_t_loss)
            epoch_v_mean = np.mean(epoch_v_loss)

            # Handle early stopping
            if self.stopping_patience > 0:
                if best_val_loss is None or epoch_v_mean < best_val_loss:
                    best_weights = flow.state_dict()
                    best_val_loss = epoch_v_mean
                    patience_counter = 0
                elif patience_counter > self.stopping_patience:
                    if self.verbose:
                        print("Early stopping training.")
                    break
                else:
                    patience_counter += 1

            if self.scheduler and scheduler is not None:
                scheduler.step(epoch_t_mean)

            if self.verbose and e % 10 == 0:
                msg = "Epoch {}/{} \tTrain loss: {}\t Val loss: {}"
                print(msg.format(e, self.epochs, epoch_t_mean, epoch_v_mean))

            train_losses.append(epoch_t_mean)
            val_losses.append(epoch_v_mean)

        # Update model weights to best
        if self.stopping_patience > 0:
            flow.load_state_dict(best_weights)

        return train_losses, val_losses



import numpy as np

import torch

from ..normalizing_flow import NormalizingFlow, NormalizingFlowFactory
from ..config_constants import *

from .odenets import WeilbachSparseODENet, FCODEnet, StrODENet, ODENet

from .ffjord.cnf import CNF
from .ffjord.container import SequentialFlow
from .ffjord.odefunc import ODEfunc

from ...models import TTuple


class ContinuousFlow(NormalizingFlow):
    """Wraps FFJORD CNF to abide to NormalizingFlow interface."""

    def __init__(self, ffjord_cnf: SequentialFlow):
        """Initialize ContinuousFlow.

        Args:
            ffjord_cnf: FFJORD CNF module.
        """
        super().__init__()
        self.model = ffjord_cnf

    def forward(self, x: torch.Tensor) -> TTuple:
        """Compute forward CNF pass.

        Args:
            x: Input data.
        Return:
            Transformed data, and jacobian determinant.
        """
        return self.model(x, reverse=False)

    def invert(self, z: torch.Tensor) -> torch.Tensor:
        """Compute inverse transform from latent to data space.

        Args:
            z: Input data from latent distribution.
        Return:
            Transformed data.
        """
        return self.model(z, reverse=True)[0]


class AdjacencyModifier:
    """Helper class to modify adjacency matrix.

    Since CNFs rely on the existence of unique ODE solutions to ensure
    invertibility, we don't need a lower triangular adjacency. We can thus
    experiment with adjacencies that yield better results.

    Empirically, adding the main diagonal to adjacency is required for good
    performance, while reflecting the adjacency does not yield significant
    performance gains.
    """

    def __init__(self, modifiers: list[str]):
        """Initialize AdjacencyModifier.

        Valid modifiers are applied in input order, and be chosen from:
        ["main_diagonal", "reflect"]

        Args:
            modified: List of modifiers to apply.
        """
        self.modifiers = modifiers

    def modify_adjacency(self, adj_mat: np.ndarray) -> np.ndarray:
        """Apply all modifiers onto adjacency matrix.

        Args:
            adj_mat: 2D binary adjacency matrix.
        Returns:
            Modified adjacency matrix.
        """
        adj_mat = np.copy(adj_mat)
        for mod in self.modifiers:
            if mod == "main_diagonal":
                adj_mat = self._mod_add_diagonal(adj_mat)
            elif mod == "reflect":
                adj_mat = self._mod_reflect(adj_mat)
            else:
                raise ValueError("Modifier not implemented.")

            # Threshold matrix to be binary.
            adj_mat = (adj_mat > 0).astype(int)

        return adj_mat

    def _mod_add_diagonal(self, adj_mat: np.ndarray) -> np.ndarray:
        """Add ones to main diagonal.

        Args:
            adj_mat: 2D binary adjacency matrix.
        Returns:
            Adjacency matrix plus ones in main diagonal.
        """
        return adj_mat + np.eye(adj_mat.shape[0])

    def _mod_reflect(self, adj_mat: np.ndarray) -> np.ndarray:
        """Reflect lower triangular values to upper triangular positions.

        Args:
            adj_mat: 2D binary adjacency matrix.
        Returns:
            Adjacency matrix with ones reflected upper triangular ones.
        """
        return adj_mat + adj_mat.T


class ContinuousFlowFactory(NormalizingFlowFactory):
    """Constructs a FFJORD continuous normalizing flow."""

    def __init__(self, config: dict):
        """Initialize ContinuousFlowFactory.

        Args:
            config: ContinuousFlow architecture parameters.
        """
        super().__init__(config)
        self.parse_config(config)

    def parse_config(self, config: dict):
        """Parse config for relevant arguments.

        Args:
            config: ContinuousFlow architecture parameters.
        """
        self.input_dim = config[INPUT_DIM]
        self.hidden_dim = config[ODENET_HID]
        self.lin_type = config[ODENET_LIN]
        self.act_type = config[ODENET_ACT]

        self.adj_mat = config[ADJ]
        self.opt_type = config[OPT_TYPE]
        self.opt_args = config[OPT_ARGS]

        self.divergence_fn = config[ODENET_DIV_FN]

        self.T = config[ODE_TIME]
        self.train_T = config[ODE_TRAIN_T]

        self.solver = config[ODE_SOLVER_TYPE]
        self.atol = config[ODE_SOLVER_ATOL]
        self.rtol = config[ODE_SOLVER_RTOL]
        self.step_size = config[ODE_SOLVER_STEP]

        self.flow_steps = config[FLOW_STEPS]

    def build_odenet(self) -> ODENet:
        """Construct neural network used to model ODE dynamics function.

        Returns:
            Neural network used to parameterize ODE dynamics.
        """
        net: ODENet | None = None
        if self.lin_type == "weilbach":
            net = WeilbachSparseODENet(
                self.input_dim,
                len(self.hidden_dim),
                self.act_type,
                self.adj_mat
            )
        elif self.lin_type == "fully_connected":
            net = FCODEnet(
                self.input_dim,
                self.hidden_dim,
                self.act_type
            )
        elif self.lin_type == "strnn":
            net = StrODENet(
                self.input_dim,
                self.hidden_dim,
                self.act_type,
                self.opt_type,
                self.opt_args,
                self.adj_mat,
            )
        else:
            raise ValueError("Unknown ODENet type.")

        return net

    def build_cnf_step(self) -> CNF:
        """Build a FFJORD continuous normalizing flow.

        TODO: Reimplement capability to regularize flows

        Returns:
            FFJORD CNF.
        """
        odenet = self.build_odenet()
        odefunc = ODEfunc(
            diffeq=odenet,
            divergence_fn=self.divergence_fn
        )
        cnf = CNF(
            odefunc=odefunc,
            T=self.T,
            train_T=self.train_T,
            solver=self.solver,
        )
        return cnf

    def _build_flow(self) -> ContinuousFlow:
        """Build a chain of continuous normalizing flows.

        Returns:
            Sequence of FFJORD CNFs wrapped in ContinuousFlow class
        """
        chain = [self.build_cnf_step() for _ in range(self.flow_steps)]
        model = SequentialFlow(chain)

        def _set(module):
            if isinstance(module, CNF):
                # Set solver settings
                module.solver = self.solver
                module.atol = self.atol
                module.rtol = self.rtol

                if self.step_size is not None:
                    module.solver_options["step_size"] = self.step_size

                # If using fixed-grid adams, restrict order to not be too high.
                if self.solver in ["fixed_adams", "explicit_adams"]:
                    module.solver_options["max_order"] = 4

        model.apply(_set)

        cnf = ContinuousFlow(model)
        return cnf



import torch
import torch.nn as nn

from torchdiffeq import odeint_adjoint as odeint  # type: ignore

from ....models import TTuple
from .odefunc import ODEfunc


class CNF(nn.Module):
    """Local copy of FFJORD's cnf.py as FFJORD is unavailable as a package.

    See: https://github.com/rtqichen/ffjord/blob/master/lib/layers/cnf.py
    """

    sqrt_end_time: torch.Tensor

    def __init__(
        self,
        odefunc: ODEfunc,
        T: float = 1.0,
        train_T: bool = False,
        solver: str = 'dopri5',
        atol: float = 1e-5,
        rtol: float = 1e-5
    ):
        """Initialize a FFJORD CNF.

        Args:
            odefunc: Dynamics function.
            T: End time of integration.
            train_T: Treats T as a trainable parameter.
            solver: ODE solver to use.
            atol: ODE solver absolute tolerance.
            rtol: ODE solver relative tolerance.
        """
        super().__init__()
        if train_T:
            t_param = nn.Parameter(torch.sqrt(torch.tensor(T)))
            self.register_parameter("sqrt_end_time", t_param)
        else:
            self.register_buffer("sqrt_end_time", torch.sqrt(torch.tensor(T)))

        self.odefunc = odefunc
        self.solver = solver
        self.atol = atol
        self.rtol = rtol
        self.test_solver = solver
        self.test_atol = atol
        self.test_rtol = rtol

    def forward(
        self,
        z: torch.Tensor,
        logpz: torch.Tensor | None = None,
        reverse: bool = False
    ) -> TTuple:
        """Compute forward pass of CNF.

        Args:
            z: Input data.
            logpz: Accumulated jacobian term.
            reverse: Flips times of integration.
        """
        if logpz is None:
            _logpz = torch.zeros(z.shape[0], 1).to(z)
        else:
            _logpz = logpz

        times = [0.0, self.sqrt_end_time * self.sqrt_end_time]
        integration_times = torch.tensor(times).to(z)

        if reverse:
            integration_times = torch.flip(integration_times, (0,))

        self.odefunc.before_odeint()

        state_t = odeint(
            self.odefunc,
            (z, _logpz),
            integration_times,
            atol=self.atol,
            rtol=self.rtol,
            method=self.solver,
            options={},
        )

        if len(integration_times) == 2:
            state_t = tuple(s[1] for s in state_t)

        z_t, logpz_t = state_t[:2]
        return z_t, logpz_t



import numpy as np
import torch


class ZukoFactorizer:
    """Implements factorization algorithm from Zuko repository."""

    def __init__(self, adjacency: np.ndarray, opt_args: dict | None = None):
        """Initialize Zuko Factorizer.

        Args:
            adjacency: Global adjacency matrix to factorize.
            opt_args: Unused.
        """
        self.adjacency = adjacency
        self.opt_args = opt_args

    def factorize(self, hidden_sizes: tuple[int, ...]) -> list[np.ndarray]:
        """Factorize adjacency matrix according to Zuko algorithm.

        Implementation adapted from: https://github.com/probabilists/zuko.

        Args:
            hidden_sizes: List of hidden widths of intermediate layers.

        Returns:
            List of all weight masks.
        """
        masks = []

        adj_mtx = torch.from_numpy(self.adjacency)
        n_outputs = self.adjacency.shape[0]

        A_prime, inv = torch.unique(adj_mtx, dim=0, return_inverse=True)
        n_deps = A_prime.sum(dim=-1)
        P = (A_prime @ A_prime.T == n_deps).double()

        indices = None

        for i, h_i in enumerate((*hidden_sizes, n_outputs)):
            if i > 0:
                # Not the first mask
                mask = P[:, indices]
            else:
                # First mask: just use rows from A
                mask = A_prime

            if sum(sum(mask)) == 0.0:
                raise ValueError("Adjacency matrix will yield null Jacobian.")

            if i < len(hidden_sizes):
                # Still on intermediate masks
                reachable = mask.sum(dim=-1).nonzero().squeeze(dim=-1)
                indices = reachable[torch.arange(h_i) % len(reachable)]
                mask = mask[indices]
            else:
                # We are at the last mask
                mask = mask[inv]

            # Need to transpose all masks to match other algorithms
            masks.append(mask.T.numpy())

        return masks



from abc import ABCMeta, abstractmethod

from math import pi

import torch
import torch.nn as nn


def standard_normal_logprob(z: torch.Tensor) -> torch.Tensor:
    """Evaluate likelihood of z under standard normal.

    Args:
        z: Input data.

    Returns:
        Log probability of data under standard normal distribution.
    """
    return -.5 * (torch.log(torch.tensor(pi) * 2) + z ** 2).sum(1)


class NormalizingFlow(nn.Module, metaclass=ABCMeta):
    """Interface for Normalizing Flows."""

    config: dict

    @abstractmethod
    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """Abstract method for NormalizingFlow forward pass."""
        pass

    @abstractmethod
    def invert(self, z: torch.Tensor) -> torch.Tensor:
        """Abstract method for NormalizingFlow reverse pass."""
        pass

    def compute_loss(self, z: torch.Tensor, jac: torch.Tensor) -> torch.Tensor:
        """Compute normalizing flow loss function.

        Computes Jacobian determinant of flow transformation summed with the
        likelihood of the transformed data under the standard normal.

        Args:
            z: Normalized data samples.
            jac: Jacobian determinant of flow transformation.

        Return:
            Normalizing flow loss of samples.
        """
        logpx = standard_normal_logprob(z) + jac
        loss = -torch.mean(logpx)
        return loss


class NormalizingFlowFactory(metaclass=ABCMeta):
    """Interface for a normalizing flow factory."""

    def __init__(self, config: dict):
        """Initialize NormalizingFlowFactory superclass.

        Args:
            config: Arguments used in flow construction.
        """
        self.config = config

    @abstractmethod
    def parse_config(self, config: dict):
        """Abstract method template for parsing model config."""
        pass

    @abstractmethod
    def _build_flow(self) -> NormalizingFlow:
        """Abstract template for concrete flow builder."""
        pass

    def build_flow(self) -> NormalizingFlow:
        """Wrap concrete flow builder to add config as an attribute."""
        flow = self._build_flow()
        flow.config = self.config

        return flow



import torch

from .normalizer import Normalizer
from ....models import TTuple


class AffineNormalizer(Normalizer):
    """Affine normalizer.

    Code taken from: https://github.com/AWehenkel/Graphical-Normalizing-Flows.
    """

    def __init__(self):
        """Initialize Affine Normalizer."""
        super().__init__()

    def forward(self, x: torch.Tensor, h: torch.Tensor) -> TTuple:
        """Compute affine transform.

        Args:
            x: Input data.
            h: Transform parameters from conditioner. Should be a (B x D x 2)
                tensor where B is batch size, D is input dimension.

        Returns:
            Transformed data and jacobian determinant.
        """
        mu = h[:, :, 0].clamp_(-5., 5.)
        sigma = torch.exp(h[:, :, 1].clamp_(-5., 2.))

        z = x * sigma + mu
        return z, sigma

    def invert(self, z: torch.Tensor, h: torch.Tensor) -> torch.Tensor:
        """Compute inverse affine transform.

        Args:
            z: Data from latent space.
            h: Transform parameters from conditioner. Should be a (B x D x 2)
                tensor where B is batch size, D is input dimension.

        Returns:
            Inverse transformed data.
        """
        mu = h[:, :, 0].clamp_(-5., 5.)
        sigma = torch.exp(h[:, :, 1].clamp_(-5., 2.))

        x = (z - mu) / sigma
        return x



from abc import ABCMeta, abstractmethod

import torch
import torch.nn as nn

from ....models import TTuple


class Normalizer(nn.Module, metaclass=ABCMeta):
    """Interface for normalizing flow normalizer."""

    @abstractmethod
    def forward(self, x: torch.Tensor, params: torch.Tensor) -> TTuple:
        """Abstract method for normalizer forward pass."""
        pass

    @abstractmethod
    def invert(self, z: torch.Tensor, params: torch.Tensor) -> torch.Tensor:
        """Abstract method for normalizer reverse pass."""
        pass




import torch
import torch.nn as nn

from UMNN import NeuralIntegral, ParallelNeuralIntegral

from .normalizer import Normalizer
from ....models import TTuple


def _flatten(sequence):
    flat = [p.contiguous().view(-1) for p in sequence]
    return torch.cat(flat) if len(flat) > 0 else torch.tensor([])


class ELUPlus(nn.Module):
    """Implement ELUPlus activation.

    Combines an ELU activation with a positive constant term.
    """

    def __init__(self):
        """Initialize ELUPlus."""
        super().__init__()
        self.elu = nn.ELU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute ELUPlus activation.

        Args:
            x: Input logits.

        Returns:
            Activated output.
        """
        return self.elu(x) + 1.05


class IntegrandNet(nn.Module):
    """Neural network which represents an integrand for use in UMNN.

    Code taken from: https://github.com/AWehenkel/Graphical-Normalizing-Flows.
    """

    def __init__(self, hidden_dim: tuple[int], n_param_per_dim: int):
        """Initialize IntegrandNet.

        Args:
            hidden_dim: Tuple of hidden widths of network.
            n_param_per_dim: Input dimension, number of parameters per
                dimension outputted by an upstream flow conditioner.
        """
        super().__init__()
        l1 = [1 + n_param_per_dim] + list(hidden_dim)
        l2 = list(hidden_dim) + [1]
        layers = []
        for h1, h2 in zip(l1, l2):
            layers += [nn.Linear(h1, h2), nn.ReLU()]
        layers.pop()
        layers.append(ELUPlus())
        self.net = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor, h: torch.Tensor) -> torch.Tensor:
        """Compute IntegrandNet forward pass.

        Args:
            x: Input data.
            h: Parameters from conditioner network.

        Returns:
            Transformed data.
        """
        nb_batch, in_d = x.shape
        x = torch.cat((x, h), 1).view(nb_batch, -1, in_d).transpose(1, 2)

        x_he = x.contiguous().view(nb_batch * in_d, -1)
        y = self.net(x_he).view(nb_batch, -1)
        return y


class MonotonicNormalizer(Normalizer):
    """Monotonic normalizer which uses UMNN to transform data.

    Code taken from: https://github.com/AWehenkel/Graphical-Normalizing-Flows.
    """

    def __init__(
        self,
        integrand_hidden: tuple[int],
        n_param_per_dim: int,
        nb_steps: int,
        solver: str
    ):
        """Initialize MonotonicNormalizer.

        Args:
            integrand_hidden: Hidden widths of integrand network.
            n_param_per_dim: Number of conditioner parameters per variable.
            nb_steps: Number of solver steps used in neural integration.
            solver: Type of solver used. Valid: ["CC", "CCParallel"]
        """
        super().__init__()

        self.integrand_net = IntegrandNet(integrand_hidden, n_param_per_dim)
        self.nb_steps = nb_steps

        solver_dict = {
            "CC": NeuralIntegral,
            "CCParallel": ParallelNeuralIntegral,
        }
        self.solver = solver_dict[solver]

    def forward(self, x: torch.Tensor, h: torch.Tensor) -> TTuple:
        """Perform monotonic transform.

        Args:
            x: Input data from data space.
            h: Parameters from conditioner.
        Returns:
            Transformed data and jacobian determinant.
        """
        x0 = torch.zeros(x.shape).to(x)
        xT = x
        z0 = h[:, :, 0]
        h = h.permute(0, 2, 1).contiguous().view(x.shape[0], -1)

        z = z0 + self.solver.apply(x0, xT, self.integrand_net,
                                   _flatten(self.integrand_net.parameters()),
                                   h, self.nb_steps)

        return z, self.integrand_net(x, h)

    def invert(self, z: torch.Tensor, h: torch.Tensor) -> torch.Tensor:
        """Perform inverse monotonic transform.

        Args:
            z: Input data from latent space.
            h: Parameters from conditioner.
        Returns:
            Inverse transformed data.
        """
        x_max = torch.ones_like(z) * 20
        x_min = -torch.ones_like(z) * 20
        z_max, _ = self.forward(x_max, h)
        z_min, _ = self.forward(x_min, h)

        for i in range(20):
            x_middle = (x_max + x_min) / 2
            z_middle, _ = self.forward(x_middle, h)
            left = (z_middle > z).float()
            right = 1 - left
            x_max = left * x_middle + right * x_max
            x_min = right * x_middle + left * x_min
            z_max = left * z_middle + right * z_max
            z_min = right * z_middle + left * z_min
        return (x_max + x_min) / 2



from .gnf_conditioner import GNFConditioner
from .made_conditioner import MADEConditioner
from .straf_conditioner import StrNNConditioner
from .conditioner import Conditioner

__all__ = [
    "Conditioner",
    "StrNNConditioner",
    "GNFConditioner",
    "MADEConditioner"
]



import numpy as np
import torch
import torch.nn as nn

from strnn import MaskedLinear
from strnn.models.model_utils import NONLINEARITIES
from .conditioner import Conditioner


class MADEConditioner(Conditioner):
    """Normalizing Flow Conditioner using MADE to compute flow parameters.

    Class is implemented using code from https://github.com/piomonti/carefl.
    This is done instead of using StrNN with "MADE" factorization to ensure
    results are directly comparable with CAREFL, but practically using
    StrNN with MADE factorization should result in the same output.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: tuple[int],
        n_out_param: int,
        act_type: str,
        num_masks: int = 1,
        natural_ordering: bool = True
    ):
        """Initialize an MADE conditioner.

        Args:
            input_dim: Dimension of input.
            hidden_dim: List of hidden widths for each layer.
            n_out_param: Number of output parameters per input variable.
            act_type: Activation function used in MADE.
            num_masks: Can be used to train ensemble over orderings/connections
            natural_ordering:
                Force natural ordering of dimensions and don't use random
                permutations.
        """
        super().__init__()
        self.input_dim = input_dim
        nout = input_dim * n_out_param
        self.nout = nout
        self.hidden_dim = hidden_dim
        assert self.nout % self.input_dim == 0

        # Define activation
        try:
            self.activation = NONLINEARITIES[act_type]
        except ValueError:
            raise ValueError(f"{act_type} is not a valid activation!")

        # define a simple MLP neural net
        self.module_list = []
        hs = [input_dim] + list(hidden_dim) + [nout]
        for h0, h1 in zip(hs, hs[1:]):
            self.module_list.extend([
                MaskedLinear(h0, h1),
                self.activation,
            ])
        self.module_list.pop()  # pop the last activation for the output layer
        self.net = nn.Sequential(*self.module_list)

        # seeds for orders/connectivities of the model ensemble
        self.natural_ordering = natural_ordering
        self.num_masks = num_masks
        self.seed = 0  # for cycling through num_masks orderings

        self.m: dict[int, np.ndarray] = {}
        self.update_masks()  # builds the initial self.m connectivity
        # note, we could also precompute the masks and cache them, but this
        # could get memory expensive for large number of masks.

    def update_masks(self):
        """Apply MADE masks to masked linear layers."""
        if self.m and self.num_masks == 1:
            return  # only a single seed, skip for efficiency
        # bool(self.m) == False if m == {} else True
        L = len(self.hidden_dim)

        # fetch the next seed and construct a random stream
        rng = np.random.RandomState(self.seed)
        self.seed = (self.seed + 1) % self.num_masks

        # sample the order of the inputs and the connectivity of all neurons
        self.m[-1] = \
            np.arange(self.input_dim) \
            if self.natural_ordering else rng.permutation(self.input_dim)
        for i in range(L):
            self.m[i] = rng.randint(self.m[i - 1].min(),
                                    self.input_dim - 1,
                                    size=self.hidden_dim[i])

        # construct the mask matrices
        masks = [self.m[i - 1][:, None] <= self.m[i][None, :]
                 for i in range(L)]
        masks.append(self.m[L - 1][:, None] < self.m[-1][None, :])

        # handle the case where nout = input_dim * k, for integer k > 1
        if self.nout > self.input_dim:
            k = int(self.nout / self.input_dim)
            # replicate the mask across the other outputs
            masks[-1] = np.concatenate([masks[-1]] * k, axis=1)

        # set the masks in all MaskedLinear layers
        layers = [layer for layer in self.net.modules()
                  if isinstance(layer, MaskedLinear)]
        for layer, m in zip(layers, masks):
            layer.set_mask(m)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute MADE conditioner forward pass.

        Args:
            x: Input data

        Returns:
            MADE conditioner output
        """
        return self.net(x).view(x.shape[0], -1, x.shape[1]).permute(0, 2, 1)



BASE_MODEL = "base_model"
INPUT_DIM = "input_dim"
ADJ = "adjacency_matrix"
ADJ_MOD = "adjacency_modifier"

OPT_TYPE = "opt_type"
OPT_ARGS = "opt_args"

FLOW_STEPS = "flow_steps"
FLOW_PERMUTE = "flow_permute"

ODENET_HID = "odenet_hid_dim"
ODENET_LIN = "odenet_lin_type"
ODENET_ACT = "odenet_act_type"

ODENET_DIV_FN = "ode_divergence_fn"
ODE_TIME = "ode_time_length"
ODE_TRAIN_T = "ode_train_T"
ODE_SOLVER_TYPE = "ode_solver_type"
ODE_SOLVER_ATOL = "ode_solver_atol"
ODE_SOLVER_RTOL = "ode_solver_rtol"
ODE_SOLVER_STEP = "ode_solver_stepsize"

COND_TYPE = "conditioner_type"
COND_HID = "conditioner_hid_dim"
COND_ACT = "conditioner_act_type"

N_PARAM_PER_VAR = "n_param_per_var"

GNF_HOT = "gnf_conditioner_hot_encoding"

NORM_TYPE = "normalizer_type"
UMNN_INT_HID = "umnn_int_hid_dim"
UMNN_INT_SOLVER = "umnn_int_solver"
UMNN_INT_STEP = "umnn_int_step"



import numpy as np

from .factorizer import AdjacencyFactorizer


class MADEFactorizer(AdjacencyFactorizer):
    """Implements factorization algorithm from MADE."""

    def __init__(self, adjacency: np.ndarray, opt_args: dict | None = None):
        """Initialize MADE factorizer.

        Factorizer and only be applied on fully autoregressive adjacencies.

        Args:
            adjacency: Unused, but must be fully lower triangular.
            opt_args: Unused.
        """
        if not np.all(adjacency == np.tril(np.ones_like(adjacency), -1)):
            raise ValueError(("MADE can only be used on a fully autoregressive"
                              " adjacency matrix."))
        self.adjacency = adjacency
        self.opt_args = opt_args

    def factorize(self, hidden_sizes: tuple[int, ...]) -> list[np.ndarray]:
        """Factorize adjacency matrix according to MADE algorithm.

        Non-random version of the MADE factorization algorithm is used.

        Args:
            hidden_sizes: List of hidden widths of intermediate layers.

        Returns:
            List of all weight masks.
        """
        m = {}
        n_layers = len(hidden_sizes)
        n_in = self.adjacency.shape[1]

        # sample the order of the inputs and the connectivity of all neurons
        m[-1] = np.arange(n_in)
        for layer in range(n_layers):
            r = [n_in - 1 - (i % n_in) for i in range(hidden_sizes[layer])]
            m[layer] = np.array(r)

        # construct the mask matrices
        masks = [m[i - 1][:, None] <= m[i][None, :] for i in range(n_layers)]
        masks.append(m[n_layers - 1][:, None] < m[-1][None, :])

        return masks



import numpy as np

import torch
import torch.nn as nn

from strnn.models. model_utils import NONLINEARITIES
from .conditioner import Conditioner


class GNFMLP(nn.Module):
    """MLP used in GNF conditioner.

    Code taken from: https://github.com/AWehenkel/Graphical-Normalizing-Flows.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: tuple[int],
        output_dim: int,
        act_type: str
    ):
        """Initialize a GNF MLP.

        Args:
            input_dim: Input dimension.
            hidden_dim: List of hidden widths.
            output_dim: Output dimension.
            act_type: Activation function used between layers.
        """
        super().__init__()

        act_fn = NONLINEARITIES[act_type]

        l1 = [input_dim] + list(hidden_dim)
        l2 = list(hidden_dim) + [output_dim]

        layers = []
        for h1, h2 in zip(l1, l2):
            layers += [nn.Linear(h1, h2), act_fn]
        layers.pop()

        self.net = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute GNFMLP forward pass."""
        return self.net(x)


class GNFConditioner(Conditioner):
    """Conditioner from Graphical Normalizing Flows.

    Code taken from: https://github.com/AWehenkel/Graphical-Normalizing-Flows.

    Uses input masking to allow neural network outputs to respect a prescribed
    adjacency. Original code contains methods to allow optimization of
    adjacency matrix based on data. While helpful, this was unrelated to our
    method so the code was removed for clarity.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: tuple[int],
        n_out_param: int,
        act_type: str,
        hot_encoding: bool,
        adj_mat: np.ndarray
    ):
        """Initialize a GNF Conditioner.

        Args:
            input_dim: Input dimension.
            hidden_dim: Tuple of hidden widths.
            n_out_param: Number of output parameters per input variable.
            act_type: Activation function used between layers.
        """
        super().__init__()
        self.adj_mat = torch.Tensor(adj_mat)

        self.input_dim = input_dim
        self.hot_encoding = hot_encoding

        in_net = input_dim * 2 if hot_encoding else input_dim

        self.embedding_net = GNFMLP(in_net, hidden_dim, n_out_param, act_type)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of GNF Conditioner.

        Note that each GNF forward pass is equivalent to calling the underlying
        GNFMLP D times.

        Args:
            x: Input data. Has shape (Batch Size x Feature Dimension).

        Returns:
            Conditioner output.
        """
        # Expand adjacency to batch size
        mask = self.adj_mat.unsqueeze(0).expand(x.shape[0], -1, -1).to(x)

        # Duplicate input by number of variables. Each variable requires its
        # own mask.
        batch = x.unsqueeze(1).expand(-1, self.input_dim, -1)

        masked_input = (batch * mask).reshape(x.shape[0] * self.input_dim, -1)

        if self.hot_encoding:
            # Append one hot encoding to inputs
            encoding = torch.eye(self.input_dim).to(x)
            encoding = encoding.unsqueeze(0).expand(x.shape[0], -1, -1)
            encoding = encoding.contiguous().view(-1, self.input_dim)

            masked_input = torch.cat((masked_input, encoding), 1)

        out = self.embedding_net(masked_input)
        out = out.view(x.shape[0], self.input_dim, -1)

        return out



from .continuous_flow import ContinuousFlowFactory, AdjacencyModifier
from .odenets import StrODENet, WeilbachSparseODENet, FCODEnet

__all__ = [
    "StrODENet",
    "WeilbachSparseODENet",
    "FCODEnet",
    "ContinuousFlowFactory",
    "AdjacencyModifier",
]



import numpy as np

from .factorizer import AdjacencyFactorizer


class GreedyFactorizer(AdjacencyFactorizer):
    """Implements greedy adjacency factorizor.

    Main factorization algorithm introduced by StrNN paper.

    Aims to maximize connections in resulting mask matrices. Factorization
    makes greedy assumption, which copies non-zero rows from the adjacency
    matrix into M2 during factorization. Empirically, this results in a
    factorization that has close to the optimal number of connections
    without high asymptotic cost.
    """

    def __init__(self, adjacency: np.ndarray, opt_args: dict | None = None):
        """Initialize a greedy algorithm mask factorizer.

        Args:
            adjacency: Global adjacency matrix to factorize.
            opt_args: Unused.
        """
        self.adjacency = adjacency
        self.opt_args = opt_args

    def factorize(self, hidden_sizes: tuple[int, ...]) -> list[np.ndarray]:
        """Factorize adjacency matrix into mask matrices.

        Factorize the given adjacency structure into per-layer masks.
        We use a recursive approach here for efficiency and simplicity.

        Args:
            hidden_sizes: List of hidden widths for each intermediate layer.

        Returns:
            List of masks in order for layers from inputs to outputs. This
            order matches how the masks are assigned to the networks in MADE.
        """
        masks: list[np.ndarray] = []
        adj_mtx = np.copy(self.adjacency)

        for layer in hidden_sizes:
            (M1, M2) = self._factorize_single_mask_greedy(adj_mtx, layer)

            # Update the adjacency structure for recursive call
            adj_mtx = M1

            # Take transpose for size: (n_inputs x n_hidden/n_output)
            masks = masks + [M2.T]

        masks = masks + [M1.T]

        return masks

    def _factorize_single_mask_greedy(
        self,
        adj_mtx: np.ndarray,
        n_hidden: int
    ) -> tuple[np.ndarray, np.ndarray]:
        """Factorize adj_mtx into M1 * M2.

        Args:
            adj_mtx: adjacency structure, n_outputs x n_inputs
            n_hidden: number of units in this hidden layer

        Returns:
            Masks (M1, M2) with the shapes:
                M1 size: (n_outputs x n_hidden)
                M2 size: (n_hidden x n_inputs)
        """
        # find non-zero rows and define M2
        A_nonzero = adj_mtx[~np.all(adj_mtx == 0, axis=1), :]
        n_nonzero_rows = A_nonzero.shape[0]
        M2 = np.zeros((n_hidden, adj_mtx.shape[1]))
        for i in range(n_hidden):
            M2[i, :] = A_nonzero[i % n_nonzero_rows]

        # find each row of M1
        M1 = np.ones((adj_mtx.shape[0], n_hidden))
        for i in range(M1.shape[0]):
            # Find indices where A is zero on the ith row
            Ai_zero = np.where(adj_mtx[i, :] == 0)[0]

            # find row using closed-form solution
            # find unique entries (rows) in j-th columns of M2 where Aij = 0
            row_idx = np.unique(np.where(M2[:, Ai_zero] == 1)[0])
            M1[i, row_idx] = 0.0

        return M1, M2



# Note: for Gaussian cases, need to construct a full 2d-by-d adj mtx
# StrNN network no longer handles Gaussian case automatically

import torch
import torch.nn.functional as F
import numpy as np
from strnn.models.strNN import StrNN
from numpy.random import binomial


SUPPORTED_DATA_TYPES = ['binary', 'gaussian']


class StrNNDensityEstimator(StrNN):
    def __init__(self,
         nin: int,
         hidden_sizes: tuple[int, ...],
         nout: int,
         opt_type: str = 'greedy',
         opt_args: dict = {'var_penalty_weight': 0.0},
         precomputed_masks: np.ndarray | None = None,
         adjacency: np.ndarray | None = None,
         activation: str = 'relu',
         data_type: str = 'binary'
    ):
        super().__init__(
            nin, hidden_sizes, nout, opt_type, opt_args,
            precomputed_masks, adjacency, activation
        )
        assert data_type in SUPPORTED_DATA_TYPES
        self.data_type = data_type

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x)

    def compute_LL(self, x, x_hat):
        """
        Compute negative log likelihood given input x and reconstructed x_hat
        """
        mu, log_sigma = x_hat[:, :self.nin], x_hat[:, self.nin:]
        z = (x - mu) * torch.exp(-log_sigma)
        log_prob_gauss = -.5 * (torch.log(self.pi * 2) + z ** 2).sum(1)
        ll = - log_sigma.sum(1) + log_prob_gauss

        return ll, z

    def get_preds_loss(self, batch):
        x = batch
        x_hat = self(x)
        assert self.data_type in SUPPORTED_DATA_TYPES

        if self.data_type == 'binary':
            # Evaluate the binary cross entropy loss
            loss = F.binary_cross_entropy_with_logits(
                x_hat, x, reduction='sum'
            ) / len(x)
        else:
            # Assume data is Gaussian if not binary
            loss = - self.compute_LL(x, x_hat)[0].sum() / len(x)

        return x_hat, loss

    def generate_sample(self, x0):
        """
        Generate a data sample using trained model
        BINARY VERSION ONLY AT THE MOMENT!!!

        @param x0: value of first data dimension
        @return: generated sample
        """
        sample = torch.from_numpy(np.zeros(self.nin))
        sample[0] = x0

        for d in range(1, self.nin):
            out = self(sample.float())
            sig = torch.nn.Sigmoid()
            out = sig(out)
            p_d = out[d]
            x_d = binomial(1, p=p_d.detach().numpy())
            sample[d] = x_d
        return sample


if __name__ == '__main__':
    A = np.ones((6, 3))
    A = np.tril(A, -1)
    model = StrNNDensityEstimator(
        nin=3,
        hidden_sizes=(6,),
        nout=6,
        opt_type='Zuko',
        opt_args={'var_penalty_weight': 0.0},
        precomputed_masks=None,
        adjacency=A,
        activation='relu')
    print(model.A)




import torch
import torch.nn as nn

from .cnf import CNF

from ....models import TTuple


class SequentialFlow(nn.Module):
    """Local copy of FFJORD's cnf.py as FFJORD is unavailable as a package.

    See: https://github.com/rtqichen/ffjord/blob/master/lib/layers/container.py
    """

    def __init__(self, layersList: list[CNF]):
        """Initialize SequentialFlow.

        Args:
            layersList: List of continuous flow steps.
        """
        super().__init__()
        self.chain = nn.ModuleList(layersList)

    def forward(self, x: torch.Tensor, reverse: bool = False) -> TTuple:
        """Compute forward pass of stacked CNF layers.

        Args:
            x: Input data.
            reverse: Directionality of flow. True indicates x->z direction.
        Returns:
            Input data transformed by sequence of flows.
        """
        if reverse:
            inds = range(len(self.chain) - 1, -1, -1)
        else:
            inds = range(len(self.chain))

        logpx = torch.zeros(x.shape[0], 1).to(x)

        for i in inds:
            x, logpx = self.chain[i](x, logpx, reverse=reverse)
        return x, -logpx



from .normalizer import Normalizer
from .affine_normalizer import AffineNormalizer
from .umnn_normalizer import MonotonicNormalizer

__all__ = [
    "Normalizer",
    "AffineNormalizer",
    "MonotonicNormalizer",
]



from setuptools import setup, find_packages

# Parse README
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="strnn-hv",
    version="1.0.1",
    author="RGKLab",
    description="PyTorch package for Structured Neural Networks.",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/rgklab/StructuredNNs",
    install_requires=[
        "torch>=2.0.0",
        "scikit-learn",
        "matplotlib",
        "torchdiffeq==0.2.3",
        "UMNN",
        "wandb"
    ],
    python_requires='>=3.11',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    packages=find_packages(
        where=".",
        exclude=["experiments*", "data*", "test*"]
    ),
)



import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from strnn.factorizers import check_masks
from strnn.factorizers import GreedyFactorizer, MADEFactorizer, ZukoFactorizer
from strnn.models.model_utils import NONLINEARITIES


OPT_MAP = {
    "greedy": GreedyFactorizer,
    "made": MADEFactorizer,
    "zuko": ZukoFactorizer,
}


class MaskedLinear(nn.Linear):
    """Weight-masked lienar layer.

    A linear neural network layer, except with a configurable binary mask
    on the weights.
    """

    mask: torch.Tensor

    def __init__(
            self,
            in_features: int,
            out_features: int,
            activation: str = 'relu'
    ):
        """Initialize MaskedLinear layer.

        Args:
            in_features: Feature dimension of input data.
            out_features: Feature dimension of output.
            activation: Unused.
        """
        super().__init__(in_features, out_features)
        # register_buffer used for non-parameter variables in the model
        self.register_buffer('mask', torch.ones(out_features, in_features))
        self.activation = activation

    def set_mask(self, mask: np.ndarray):
        """Store mask for use during forward pass.

        Arg:
            mask: Mask applied to weight matrix.
        """
        self.mask.data.copy_(torch.from_numpy(mask.astype(np.uint8).T))

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Compute forward pass for MaskedLinear.

        Applies weight mask to weight matrix to block desired connections.

        Args:
            input: Input data.

        Returns:
            MaskedLinear output.
        """
        # Note: * is element-wise multiplication in numpy
        return F.linear(input, self.mask * self.weight, self.bias)


class StrNN(nn.Module):
    """Main neural network class that implements a Structured Neural Network.

    Can also become a MADE or Zuko masked NN by specifying the opt_type flag
    """

    def __init__(
        self,
        nin: int,
        hidden_sizes: tuple[int, ...],
        nout: int,
        opt_type: str = 'greedy',
        opt_args: dict = {'var_penalty_weight': 0.0},
        precomputed_masks: np.ndarray | None = None,
        adjacency: np.ndarray | None = None,
        activation: str = 'relu'
    ):
        """Initialize a Structured Neural Network (StrNN).

        Args:
            nin: input dimension
            hidden_sizes: list of hidden layer sizes
            nout: output dimension
            opt_type: optimization type: greedy, zuko, MADE
            opt_args: additional optimization algorithm params
            precomputed_masks: previously stored masks, use directly
            adjacency: the adjacency matrix, nout by nin
            activation: activation function to use in this NN
        """
        super().__init__()

        # Set parameters
        self.nin = nin
        self.hidden_sizes = hidden_sizes
        self.nout = nout

        # Define activation
        try:
            self.activation = NONLINEARITIES[activation]
        except ValueError:
            raise ValueError(f"{activation} is not a valid activation!")

        # Define StrNN network
        self.net_list = []
        hs = [nin] + list(hidden_sizes) + [nout]  # list of all layer sizes
        for h0, h1 in zip(hs, hs[1:]):
            self.net_list.extend([
                MaskedLinear(h0, h1),
                self.activation
            ])

        # Remove the last activation for the output layer
        self.net_list.pop()
        self.net = nn.Sequential(*self.net_list)

        # Load adjacency matrix
        self.opt_type = opt_type.lower()
        self.opt_args = opt_args

        if adjacency is not None:
            self.A = adjacency
        else:
            if self.opt_type == "made":
                # Initialize adjacency structure to fully autoregressive
                warnings.warn(("Adjacency matrix is unspecified, defaulting to"
                               " fully autoregressive structure."))
                self.A = np.tril(np.ones((nout, nin)), -1)
            else:
                raise ValueError(("Adjacency matrix must be specified if"
                                  "factorizer is not MADE."))

        # Setup adjacency factorizer
        try:
            self.factorizer = OPT_MAP[self.opt_type](self.A, self.opt_args)
        except ValueError:
            raise ValueError(f"{opt_type} is not a valid opt_type!")

        self.precomputed_masks = precomputed_masks

        # Update masks
        self.update_masks()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Propagates the input forward through the StrNN network.

        Args:
            x: Input of size (sample_size by data_dimensions)
        Returns:
            Output of size (sample_size by output_dimensions)
        """
        return self.net(x)

    def update_masks(self):
        """Update masked linear layer masks to respect adjacency matrix."""
        if self.precomputed_masks is not None:
            # Load precomputed masks if provided
            masks = self.precomputed_masks
        else:
            masks = self.factorizer.factorize(self.hidden_sizes)

        self.masks = masks
        assert check_masks(masks, self.A), "Mask check failed!"

        # For when each input produces multiple outputs
        # e.g. each x_i gives mean and variance for Gaussian density estimation
        if self.nout != self.A.shape[0]:
            # Then nout should be an exact multiple of nin
            assert self.nout % self.nin == 0
            k = int(self.nout / self.nin)
            # replicate the mask across the other outputs
            masks[-1] = np.concatenate([masks[-1]] * k, axis=1)

        # Set the masks in all MaskedLinear layers
        layers = [m for m in self.net.modules() if isinstance(m, MaskedLinear)]
        for layer, mask in zip(layers, self.masks):
            layer.set_mask(mask)



from abc import ABCMeta, abstractmethod

import torch
import torch.nn as nn


class Conditioner(nn.Module, metaclass=ABCMeta):
    """Interface for Normalizing Flow Conditioner."""

    input_dim: int

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Abstract method for forward pass."""
        pass



import torch
import torch.nn as nn

from ....models import TTuple


class DivergenceFunction():
    """Implements computations for divergence function."""

    def __init__(self, divergence_fn: str):
        """Initialize divergence function."""
        assert divergence_fn in ("brute_force", "approximate")
        self.divergence_fn = divergence_fn

    def get_divergence(
        self,
        dy: torch.Tensor,
        y: torch.Tensor,
        e: torch.Tensor
    ) -> torch.Tensor:
        """Compute divergence function."""
        if self.divergence_fn == "brute_force":
            return self.divergence_bf(dy, y)
        elif self.divergence_fn == "approximate":
            return self.divergence_approx(dy, y, e)
        else:
            raise ValueError()

    def divergence_bf(self, dx: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Compute brute-force divergence."""
        sum_diag = torch.zeros(dx.shape[0]).to(y)
        for i in range(y.shape[1]):
            grad = torch.autograd.grad(dx[:, i].sum(), y, create_graph=True)
            sum_diag += grad[0].contiguous()[:, i].contiguous()
        return sum_diag.contiguous()

    def divergence_approx(
        self,
        f: torch.Tensor,
        y: torch.Tensor,
        e: torch.Tensor | None = None
    ) -> torch.Tensor:
        """Compute approximate divergence."""
        e_dzdx = torch.autograd.grad(f, y, e, create_graph=True)[0]
        e_dzdx_e = e_dzdx * e
        approx_tr_dzdx = e_dzdx_e.view(y.shape[0], -1).sum(dim=1)
        return approx_tr_dzdx


class ODEfunc(nn.Module):
    """Local copy of FFJORD's odefunc.py as FFJORD is unavailable as a package.

    https://github.com/rtqichen/ffjord/blob/master/lib/layers/odefunc.py
    """

    def __init__(self, diffeq: nn.Module, divergence_fn: str = "approximate"):
        """Initialize ODEFunc.

        Args:
            diffeq: Network that models ODE dynamics.
            divergence_fn: Specifies ODE divergence function.
        """
        super().__init__()

        self.diffeq = diffeq
        self.div_fn = DivergenceFunction(divergence_fn)

        self._e: torch.Tensor | None = None

    def before_odeint(self, e: torch.Tensor | None = None):
        """Run actions prior to ODE integration."""
        self._e = e

    def forward(self, t: torch.Tensor, states: TTuple) -> TTuple:
        """Compute forward pass for ODE dynamics."""
        assert len(states) == 2
        y = states[0]

        # convert to tensor
        if not torch.is_tensor(t):
            t = torch.tensor(t).to(y)
        t = t.type_as(y)

        batchsize = y.shape[0]

        # Sample and fix the noise.
        if self._e is None:
            self._e = torch.randn_like(y).to(y)

        with torch.set_grad_enabled(True):
            y.requires_grad_(True)
            t.requires_grad_(True)
            dy = self.diffeq(t, y)

            # Hack for 2D data to use brute force divergence computation.
            if not self.training and dy.view(dy.shape[0], -1).shape[1] == 2:
                divergence_bf = self.div_fn.divergence_bf
                divergence = divergence_bf(dy, y).view(batchsize, 1)
            else:
                divergence = self.div_fn.get_divergence(dy, y, self._e)
                divergence = divergence.view(batchsize, 1)

        return (dy, -divergence)



import numpy as np

import torch

from strnn import StrNN
from .conditioner import Conditioner


class StrNNConditioner(Conditioner):
    """Normalizing Flow Conditioner using StrNN to compute flow parameters."""

    def __init__(
        self,
        input_dim: int,
        hidden_dim: tuple[int],
        n_out_param: int,
        act_type: str,
        adj_mat: np.ndarray,
        opt_type: str,
        opt_args: dict
    ):
        """Initialize an StrNN conditioner.

        Args:
            input_dim: Dimension of input.
            hidden_dim: List of hidden widths for each layer.
            n_out_param: Number of output parameters per input variable.
            act_type: Activation function used in StrNN.
            adj_mat: 2D binary adjacency matrix.
            out_type: Matrix factorization used by StrNN.
            opt_args: Dictionary of arguments to pass to factorization.
        """
        super().__init__()
        self.input_dim = input_dim
        self.strnn = StrNN(input_dim, hidden_dim, input_dim * n_out_param,
                           opt_type, opt_args, None, adj_mat, act_type)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute forward pass for StrNN Conditioner.

        Output is reshaped so all output that corresponds to a variable is
        grouped in the last dimension. Care must be taken so that outputs which
        do not respect the independence relations for a variable are grouped.

        The StrNN stacks blocks of outputs, e.g., [1 2 3 4 1 2 3 4] instead of
        [1 1 2 2 3 3 4 4], and the reshape operation must follow this.
        Args:
            x: Input data
        Returns:
            Parameters for use in Normalizer.
        """
        return self.strnn(x).view(x.shape[0], -1, x.shape[1]).permute(0, 2, 1)
